﻿'Copyright © exc-jdbi 2020
'www.vb-paradise.de/index.php/User/23082-exc-jdbi/
'© PwKeySetHolder 2020 V1

'Siehe beiligende Lizenz

Option Strict On
Option Explicit On

Imports System.IO
Imports System.Security.Cryptography
Imports exc.jdbi.Cryptographie.PwKeySetHolders

'HINWEIS:
'Unter Verweis "System.Security" einbinden.
'Unter Verweis "PwKeySetHolder" einbinden.

'Hinweis
'Nur für Windowsplattformen
'https://docs.microsoft.com/de-de/dotnet/api/system.security.cryptography.protecteddata?view=netframework-4.7.1#remarks


'Zeigt ein Beispiel für das Versenden von vertraulichen
'Daten übers Internet. Eignet sich sehr gut für sehr grosse
'verkryptete Datenpackete dessen Passwort unmittelbar vor 
'versenden übers Internet auch noch verkryptet wird.

' ******* ******* ******* ******* ******* ******* ******* ******* 

'- Das Cipherpacket ist schon verkryptet und das Passwort
'  wird nun auch noch verkryptet.
'- Jetzt werden die verkrypteten Daten übers Internet an
'  Bob zugestellt. Bob nimmt sich das verkryptete Passwort
'  und entkryptet es mit den VereinbahrungsKeys (Diffie-Hellman).
'- Nun kann Bob auch noch das grosse Cipherpacket entkrypten
'  und daraus das ursprünglich erstellte KeySet extrahieren.


Public Module DiffieHellmanTest
  Public Sub MainTestDH()
    Console.WriteLine("DiffieHellmanTest")

    Dim sw = New Stopwatch
    For i = 0 To 4
      sw.Restart()
      Alice.Main()
      Console.WriteLine("Test{0} DH t = {1}ms", i, sw.ElapsedMilliseconds)
    Next
    Console.WriteLine()
  End Sub
End Module


'Eine Möglichkeit mit Diffie-Hellman-Schlüsselaustausch

'Aus der MSDN-Dokumentation
'https://docs.microsoft.com/de-de/dotnet/api/system.security.cryptography.ecdiffiehellmancng?view=netframework-4.8


Public Module Alice
  Public AlicePublicKey() As Byte

  Public Sub Main()
    'Instanz Alice
    Using alice As New ECDiffieHellmanCng()

      'Ableitungsfunktion wählen.
      alice.KeyDerivationFunction =
        ECDiffieHellmanKeyDerivationFunction.Hash

      'HashAlgo wählen.
      alice.HashAlgorithm = CngAlgorithm.Sha256

      'Den PublicKey in eine Globalvariable, damit die
      'für die Klasse Bob auch zugänglich ist. (fiktiv)
      AlicePublicKey = alice.PublicKey.ToByteArray()

      'die Instanz von Bob
      Dim bob As New Bob()

      'Versendedaten bekannt geben
      Dim dispatch As DispatchData = Nothing

      'Kurz prüfen, ob beide PublicKeys nicht gleich sind.
      '  >> Dürfen auf keinen Fall gleich sein!
      If Not AlicePublicKey.SequenceEqual(bob.BobPublicKey) Then

        'Schlüsselmateria für Alice herleiten
        Dim aliceKey = alice.DeriveKeyMaterial(
          CngKey.Import(bob.BobPublicKey,
          CngKeyBlobFormat.EccPublicBlob)
        )

        'Ein paar Variablen bekannt geben
        Dim secretmsg, cipherpackage As Byte()

        'Schutzart kekannt geben
        Dim protection As DataProtectionScope
        protection = If((Rand.Next And 1) = 0,
                         DataProtectionScope.CurrentUser,
                         DataProtectionScope.LocalMachine)

        'PwKeySetHolder instanzieren
        Using pks = New PwKeySetHolder

          'Neuer KeySet generieren
          pks.CreateNewKeySet(10, 15, protection)

          'PlainKeySet extrahieren.
          Dim KeyAsPlain = pks.ToPwKeySet

          'Für den Test eine Kopie an GlobalKeySet
          GlobalKeySet = KeyAsPlain.Select(Function(bb) bb.ToArray).ToArray


          'Versendedaten erstellen hier ohne USING
          'da beide Referenzwerte noch gebraucht werden.
          dispatch = pks.ToDispatchData
          'Das Passwort zuweisen, denn es wird von
          'Alice nachher verkryptet
          secretmsg = dispatch.Plain_Password

          'ChiperData zu den CipherMessages zuweisen
          'es wird nicht mehr verkryptet, und kann einfach
          'übers Internet versendet werden.
          cipherpackage = dispatch.ChiperData
        End Using

        'Senden zum Verkrypten
        Dim encmsgiv = Send(aliceKey, secretmsg)

        'secretmsg kann zurückgesetzt werden
        Array.Clear(secretmsg, 0, secretmsg.Length)

        'Und hier ist nun das verkryptete Passwort ...
        Dim encryptmsg = encmsgiv.Item1

        '.... plus die Vereinbarungs-IV ...
        Dim iv = encmsgiv.Item2

        '... plus die message = ChiperData.
        bob.Receive(encryptmsg, iv, cipherpackage)
      End If
      If dispatch IsNot Nothing Then
        dispatch.Dispose()
      End If
    End Using
  End Sub
  Private Function Send(key() As Byte, secretmsg() As Byte) As (Byte(), Byte())
    If key Is Nothing OrElse key.Length < 1 Then _
          Throw New ArgumentException(NameOf(key))
    If secretmsg Is Nothing OrElse secretmsg.Length < 1 Then _
          Throw New ArgumentException(NameOf(secretmsg))

    Dim iv, encryptmsg As Byte()
    Using aes As New AesCryptoServiceProvider
      aes.Key = key : aes.GenerateIV() : iv = aes.IV
      Using ciphertext As New MemoryStream
        Using cs As New CryptoStream(
          ciphertext,
          aes.CreateEncryptor(),
          CryptoStreamMode.Write
          )
          cs.Write(secretmsg, 0, secretmsg.Length)
        End Using
        encryptmsg = ciphertext.ToArray()
      End Using
    End Using
    Return (encryptmsg, iv)
  End Function

End Module


Public Class Bob
  Public BobPublicKey() As Byte
  Private ReadOnly BobKey() As Byte
  Public Sub New()
    Using bob As New ECDiffieHellmanCng()
      'Ableitungsfunktion wählen.
      bob.KeyDerivationFunction = ECDiffieHellmanKeyDerivationFunction.Hash

      'HashAlgo definieen
      bob.HashAlgorithm = CngAlgorithm.Sha256

      'Bobs PublicKey Öffentlich setzen
      Me.BobPublicKey = bob.PublicKey.ToByteArray()

      'Bobs Schlüsselmaterial herleiten
      Me.BobKey = bob.DeriveKeyMaterial(CngKey.Import(Alice.AlicePublicKey,
                                        CngKeyBlobFormat.EccPublicBlob))
    End Using
  End Sub
  Public Sub Receive(encryptmsg() As Byte,
                     iv() As Byte,
                     cipherpackage() As Byte)
    Dim password() As Byte
    'AES-Entcryptung
    Using aes As New AesCryptoServiceProvider()
      aes.Key = Me.BobKey : aes.IV = iv
      ' Decrypt the message
      Using plaintext As New MemoryStream()
        Using cs As New CryptoStream(
          plaintext,
          aes.CreateDecryptor(),
          CryptoStreamMode.Write
          )
          cs.Write(encryptmsg, 0, encryptmsg.Length)
        End Using

        'Nun ist das Passwort wieder als Klartext
        password = plaintext.ToArray()
      End Using

      'Variable für den KeySet bekannt geben.
      Dim keyset()() As Byte

      'PwKeySetHolder-Instanz erstellen
      Using pks = New PwKeySetHolder
        'Ein neue DispatchData-Instanz erstellen und
        'das Passwort plus die CipherData einfliessen
        'lassen. 
        Using dispatch = New DispatchData With
          {
            .Plain_Password = password,
            .ChiperData = cipherpackage
          }
          'DispatchDaten importieren
          pks.ImportDispatchData(dispatch)
          'PlainKeySet extrahieren
          keyset = pks.ToPwKeySet
          'Stop 'Der Schlüssel ist nun bei BOB
          Debug.Assert(Me.Equality(keyset, GlobalKeySet))
        End Using
      End Using
    End Using
  End Sub
  Private Function Equality(b1() As Byte, b2() As Byte) As Boolean
    Return b1.SequenceEqual(b2)
  End Function
  Private Function Equality(b1()() As Byte, b2()() As Byte) As Boolean
    If Not b1.Length = b2.Length Then Return False
    For i = 0 To b1.Length - 1
      If Not Me.Equality(b1(i), b2(i)) Then
        Return False
      End If
    Next
    Return True
  End Function
End Class

Friend Module GlobalTestHelpers
  'Dient nur für den Test, um den erzeugten KeySet
  'global überall zugänglich zu machen.
  'Wird dann mit dem erzeugten KeySet von Bob verglichen
  Friend GlobalKeySet()() As Byte
End Module
